# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         testLint.py
# Purpose:      Controller for all lint based testing.
#
# Authors:      Christopher Ariza
#               Michael Scott Cuthbert
#
# Copyright:    Copyright © 2009-2010, 2015 Michael Scott Cuthbert and the music21 Project
# License:      LGPL or BSD, see license.txt
# ------------------------------------------------------------------------------

# this requires pylint to be installed and available from the command line
import argparse
import pathlib

from . import common
from . import commonTest

try:
    from pylint.lint import Run as pylintRun
except ImportError:
    pylintRun = None


# see feature list here:
# http://docs.pylint.org/features.html

# W0511:	Used when a warning note as FIXME or XXX is detected.
# W0404:	Reimport %r (imported line %s) Used when a module is reimported multiple times.

# we do this all the time in unit tests
# R0201:	Method could be a function Used when a method doesn't
#           use its bound instance, and so could be written as a function.
# R0904:	Too many public methods (%s/%s) Used when class has too many public methods,
#            try to reduce this to get a more simple (and so easier to use) class.
# E1101:	%s %r has no %r member Used when a variable is accessed for an non-existent member.
# R0914:	Too many local variables (%s/%s) Used when a function or method has
#           too many local variables.
# many of our test use many local variables
# R0903:	Too few public methods (%s/%s) Used when class has too few public methods,
#                  so be sure it's really worth it.
# R0911:	Too many return statements (%s/%s) Used when a function or method has
#                  too many return statement, making it hard to follow.
def main(fnAccept=None, strict=False):
    '''
    `fnAccept` is a list of one or more files to test.  Otherwise runs all.
    '''
    poolSize = common.cpus()

    if pylintRun is None:
        print("make sure that 'sudo pip3 install pylint' is there. exiting.")
        return

    mg = commonTest.ModuleGather()

    fnPathReject = [
                    '/ext/',
                    ]

    disable_unless_strict = [
                'too-many-statements', # someday
                'too-many-arguments', # definitely! but takes too long to get a fix now...
                'too-many-public-methods', # maybe, look
                'too-many-branches', # yes, someday
                'too-many-lines',    # yes, someday.
                'too-many-return-statements', # we'll see
                'too-many-instance-attributes', # maybe later
                'inconsistent-return-statements', # would be nice
                'protected-access', # this is an important one, but for now we do a lot of
                           # x = copy.deepcopy(self); x._volume = ... which is not a problem...
                           # also, test suites need to be exempt.
                'keyword-arg-before-vararg', # a good thing to check for new code, but
                           # requires rewriting function signatures in old code

    ]
    disable = [  # These also need to be changed in .pylintrc
                'arguments-differ', # -- no -- should be able to add additional arguments so long
                    # as initial ones are the same.
                'multiple-imports', # import os, sys -- fine...
                'redefined-variable-type', # would be good, but currently
                # lines like: if x: y = note.Note() ; else: y = note.Rest()
                # triggers this, even though y doesn't change.
                'no-else-return', # these are unnessary but can help show the flow of thinking.
                'cyclic-import', # we use these inside functions when there's a deep problem.
                'unnecessary-pass', # nice, but not really a problem...
                'locally-disabled', # test for this later, but hopefully will know what
                            # they're doing
                'consider-using-get', # if it can figure out that the default value is something
                            # simple, we will turn back on, but until then, no.
                'chained-comparison', # sometimes simpler that way
                # 'duplicate-code', # needs to ignore strings -- keeps getting doctests...
                'too-many-ancestors', # -- 8 is okay.
                'abstract-class-instantiated', # this trips on the fractions.Fraction() class.
                'fixme', # known...
                'superfluous-parens', # nope -- if they make things clearer...
                'no-member', # important, but too many false positives
                'too-many-locals',   # no
                'bad-whitespace', # maybe later, but "bad" isn't something I necessarily agree with
                'bad-continuation',  # never remove -- this is a good thing many times.
                'unpacking-non-sequence', # gets it wrong too often.
                'too-many-boolean-expressions', # AbstractDiatonicScale.__eq__ shows how this
                    # can be fine...
                'misplaced-comparison-constant', # sometimes 2 < x is what we want
                'unsubscriptable-object', # unfortunately, thinks that Streams are unsubscriptable.
                'consider-iterating-dictionary', # sometimes .keys() is a good test against
                    # changing the dictionary size while iterating.

                'invalid-name',      # these are good music21 names; fix the regexp instead...
                'no-self-use',       # maybe later
                'too-few-public-methods', # never remove or set to 1

                'trailing-whitespace',  # should ignore blank lines with tabs
                'trailing-newlines', # just because something is easy to detect doesn't make it bad.

                'missing-docstring',    # gets too many well-documented properties
                'star-args', # no problem with them...
                'unused-argument',
                'import-self', # fix is either to get rid of it or move away many tests...

                'simplifiable-if-statement', # NO! NO! NO!
                #  if (x or y and z and q): return True, else: return False,
                #      is a GREAT paradigm -- over "return (x or y and z and q)" and
                #      assuming that it returns a bool...  it's no slower than
                #      the simplification and it's so much clearer.
                'consider-using-enumerate', # good when i used only once, but
                    # x[i] = y[i] is a nice paradigm, even if one can be simplified out.
               ]
    if not strict:
        disable = disable + disable_unless_strict

    goodnameRx = {'argument-rgx': r'[a-z_][A-Za-z0-9_]{2,30}$',
                  'attr-rgx': r'[a-z_][A-Za-z0-9_]{2,30}$',
                  'class-rgx': r'[A-Z_][A-Za-z0-9_]{2,30}$',
                  'function-rgx': r'[a-z_][A-Za-z0-9_]{2,30}$',
                  'method-rgx': r'[a-z_][A-Za-z0-9_]{2,30}$',
                  'module-rgx': r'(([a-z_][a-zA-Z0-9_]*)|([A-Z][a-zA-Z0-9]+))$',
                  'variable-rgx': r'[a-z_][A-Za-z0-9_]{2,30}$',
                  }

    maxArgs = 7 if not strict else 5
    maxBranches = 20 if not strict else 10

    cmd = ['--output-format=parseable',
           r'--dummy-variables-rgx="_$|dummy|unused|i$|j$|junk|counter"',
           '--docstring-min-length=3',
           '--ignore-docstrings=yes',
           '--min-similarity-lines=8',
           '--max-args=' + str(maxArgs),  # should be 5 later, but baby steps
           '--bad-names="foo,shit,fuck,stuff"', # definitely allow "bar" for barlines
           '--reports=n',
           '--max-branches=' + str(maxBranches),
           '-j ' + str(poolSize), # multiprocessing!
           r'--ignore-long-lines="converter\.parse"', # some tiny notation...
           '--max-line-length=100', # tada
           ]
    for gn, gnv in goodnameRx.items():
        cmd.append('--' + gn + '="' + gnv + '"')

    for pyLintId in disable:
        cmd.append('--disable=%s' % pyLintId)

    # add entire package
    acceptable = []
    for fp in mg.modulePaths:
        rejectIt = False
        for rejectPath in fnPathReject:
            if rejectPath in fp:
                rejectIt = True
                break
        if rejectIt:
            continue
        if fnAccept:
            rejectIt = True
            for acceptableName in fnAccept:
                if acceptableName in fp:
                    rejectIt = False
                    break
            if rejectIt:
                continue

        acceptable.append(fp)

    cmdFile = cmd + acceptable
    # print(' '.join(cmdFile))
    # print(fp)
    try:
        pylintRun(cmdFile, exit=False)
    except TypeError:
        pylintRun(cmdFile, do_exit=False) # renamed in recent versions


def arg_run():
    parser = argparse.ArgumentParser(
        description='Run pylint on the project according to style guide.')
    parser.add_argument('files', metavar='filename', type=str, nargs='*',
                        help='Files to parse (default nearly all)')
    parser.add_argument('--strict', action='store_true',
                        help='Run the file in strict mode')
    args = parser.parse_args()
    # print(args.files)
    # print(args.strict)
    files = args.files if args.files else None
    if files:
        sfp = common.source_file_path()
        files = [pathlib.Path(f).relative_to(sfp) for f in files]
    main(files, args.strict)


if __name__ == '__main__':
    arg_run()
    # main(sys.argv[1:])
    #     if len(sys.argv) >= 2:
    #         test.main(sys.argv[1:], restoreEnvironmentDefaults=True)
    #     else:
    #         test.main(restoreEnvironmentDefaults=True)
    #

# -----------------------------------------------------------------------------
# eof



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         coverageProject.py
# Purpose:      Starts Coverage w/ default arguments
#
# Authors:      Christopher Ariza
#               Michael Scott Cuthbert
#
# Copyright:    Copyright © 2014-15 MIT DH Project
#               forked from music21, Copyright © 2014-15 Michael Scott Cuthbert
#               and the music21 Project
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------
import sys

omit_modules = [
                ]

# THESE ARE NOT RELEVANT FOR coveralls.io -- edit .coveragerc to change that
exclude_lines = [
                r'.*main_test\(',
                r'.*#\s*pragma:\s*no cover.*',
                r'class .*External.*',
                r'class .*Slow.*',
                ]


def get_coverage(overrideVersion=False):
    if overrideVersion or sys.version_info.minor == 5:
        # run on Py 3.5 -- to get Py 3.6/3.7 timing...
        try:
            import coverage
            cov = coverage.coverage(omit=omit_modules)
            for e in exclude_lines:
                cov.exclude(e, which='exclude')
            cov.start()
        except ImportError:
            cov = None
    else:
        cov = None
    return cov


def stop_coverage(cov):
    if cov is not None:
        cov.stop()
        cov.save()



# -*- coding: utf-8 -*-
__all__ = [
    'coverageProject',
    'commonTest',
    'testExternal',
    'testPerformance',
    'testRunner',
    'timeGraphs',
    'testStream',
    ]

_DOC_IGNORE_MODULE_OR_PACKAGE = True

# ------------------------------------------------------------------------------
# eof



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         multiprocesssTest.py
# Purpose:      Controller for all tests in project run concurrently.
#
# Authors:      Michael Scott Cuthbert
#
# Copyright:    Copyright © 2012-18 MIT DH Lab
#               forked from music21 Copyright © 2012-15
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------
'''
Multiprocess testing.  Tests all doctests and Test unittest objects in all
modules that are imported when running "import music21".  Runs threads on
each core of a multicore system unless there are more than 2 cores, in which
case it runs on n-1 cores.

N.B. this gets a slightly different set of modules than test/test.py does
because the `imp` module is not available for threaded processing.  Running
both modules gives great coverage of just about everything -- do that before
building a new release.

Run testDocumentation after this.
'''
import collections
import multiprocessing
import os
import sys
import tempfile
import time
import unittest

from . import common
from . import testRunner
from . import commonTest

ModuleResponse = collections.namedtuple('ModuleResponse',
                    'returnCode fp moduleName success testRunner '
                    + 'errors failures testsRun runTime')
ModuleResponse.__new__.__defaults__ = (None,) * len(ModuleResponse._fields)


# ------------------------------------------------------------------------------

def run_one_module_without_imp(arguments):
    modGath, fp = arguments # modGather object, filepath
    verbosity = False
    timeStart = time.time()
    moduleObject = modGath.get_module_without_imp(fp)
    globs = common.import_main_module().__dict__.copy()

    # environLocal.printDebug('running %s \n' % fp)
    if moduleObject == 'skip':
        success = '%s is skipped \n' % fp
        # environLocal.printDebug(success)
        return ModuleResponse('Skipped', fp, success)
    elif moduleObject == 'notInTree':
        success = ('%s is in the directory but not imported in project. Skipped -- fix!' %
                   modGath._getNamePeriod(fp))
        # environLocal.printDebug(success)
        return ModuleResponse("NotInTree", fp, success)


    try:
        moduleName = modGath._getName(fp)

        s1 = commonTest.default_doctest_suite()

        # get Test classes in moduleObject
        if not hasattr(moduleObject, 'Test'):
            pass
            # environLocal.printDebug('%s has no Test class' % moduleObject)
        else:
            s2 = unittest.defaultTestLoader.loadTestsFromTestCase(moduleObject.Test)
            s1.addTests(s2)

        try:
            s3 = commonTest.default_doctest_suite(moduleObject, globs=globs)
            s1.addTests(s3)
        except ValueError:
            # environLocal.printDebug('%s cannot load Doctests' % moduleObject)
            pass

        testRunner.fixDoctests(s1)


        # environLocal.printDebug('running Tests...\n')
        runner = commonTest.ProjectTestRunner(verbosity=verbosity)
        try:
            testResult = runner.run(s1)

            # need to make testResult pickleable by removing the instancemethod parts...
            errors = []
            for e in testResult.errors:
                errors.append(e[1])
            failures = []
            for f in testResult.failures:
                failures.append(f[1])
            runTime = round(10 * (time.time() - timeStart)) / 10.0
            return ModuleResponse("TestsRun", fp, moduleName, testResult.wasSuccessful(),
                                  str(testResult), errors, failures, testResult.testsRun, runTime)
        except Exception as excp: # pylint: disable=broad-except
            # environLocal.printDebug('*** Exception in running %s: %s...\n' % (moduleName, excp))
            return ModuleResponse("TrappedException", fp, moduleName, None, str(excp))
    except Exception as excp: # pylint: disable=broad-except
        # environLocal.printDebug('*** Large Exception in running %s: %s...\n' % (fp, excp))
        return ModuleResponse("LargeException", fp, None, None, str(excp))


def main(test_group=None):
    '''
    Run all tests. test_group is not used (always the test suite)
    '''
    normalStdError = sys.stderr

    timeStart = time.time()
    poolSize = common.cpus()

    print('Creating %d processes for multiprocessing' % poolSize)

    modGather = commonTest.ModuleGather(useExtended=True)

    maxTimeout = 200
    pathsToRun = modGather.modulePaths # [30:60]

    pool = multiprocessing.Pool(processes=poolSize) # @UndefinedVariable # pylint: disable=not-callable

    # imap returns the results as they are completed.  Since the number of files is small,
    # the overhead of returning is outweighed by the positive aspect of getting results immediately
    # unordered says that results can RETURN in any order; not that they'd be pooled out in any
    # order.
    modGather_with_fp = [(modGather, fp) for fp in pathsToRun]
    res = pool.imap_unordered(run_one_module_without_imp, modGather_with_fp)

    continueIt = True
    timeouts = 0
    eventsProcessed = 0
    summaryOutput = []

    while continueIt is True:
        try:
            newResult = res.next(timeout=1)
            if timeouts >= 5:
                print("")
            if newResult is not None:
                if newResult.moduleName is not None:
                    mn = str(newResult.moduleName)
                    mn = mn.replace('___init__', '')
                    mn = mn.replace('_', '.')
                else:
                    mn = ""
                rt = newResult.runTime
                if rt is not None:
                    rt = round(newResult.runTime * 10) / 10
                    if not newResult.errors and not newResult.failures:
                        print("\t\t\t\t{0}: {1} tests in {2} secs".format(
                                            mn,
                                            newResult.testsRun,
                                            rt))
                    else:
                        print("\t\t\t\t{0}: {1} tests, {2} errors {3} failures in {4} secs".format(
                                            mn,
                                            newResult.testsRun,
                                            len(newResult.errors),
                                            len(newResult.failures),
                                            rt))
            timeouts = 0
            eventsProcessed += 1
            summaryOutput.append(newResult)
        except multiprocessing.TimeoutError: # @UndefinedVariable
            timeouts += 1
            if timeouts == 5 and eventsProcessed > 0:
                print("Delay in processing, seconds: ", end="")
            elif timeouts == 5:
                print("Starting first modules, should take 5-10 seconds: ", end="")

            if timeouts % 5 == 0:
                print(str(timeouts) + " ", end="", flush=True)
            if timeouts > maxTimeout and eventsProcessed > 0:
                print("\nToo many delays, giving up...", flush=True)
                continueIt = False
                printSummary(summaryOutput, timeStart, pathsToRun)
                pool.close()
                exit()
        except StopIteration:
            continueIt = False
            pool.close()
            pool.join()
        except Exception as excp: # pylint: disable=broad-except
            eventsProcessed += 1
            exceptionLog = ModuleResponse("UntrappedException", None, "%s" % excp)
            summaryOutput.append(exceptionLog)

    sys.stderr = normalStdError
    printSummary(summaryOutput, timeStart, pathsToRun)


def printSummary(summaryOutput, timeStart, pathsToRun):
    outStr = ''
    summaryOutputTwo = [i[1] for i in summaryOutput]
    for fp in pathsToRun:
        if fp not in summaryOutputTwo:
            failLog = ModuleResponse("NoResult", fp)
            summaryOutput.append(failLog)

    totalTests = 0

    skippedSummary = []
    successSummary = []
    errorsFoundSummary = []
    otherSummary = []
    for moduleResponse in summaryOutput:
        print(moduleResponse)
        if moduleResponse.returnCode == 'Skipped':
            skippedSummary.append("Skipped: %s" % moduleResponse.fp)
        elif moduleResponse.returnCode == 'NoResult':
            otherSummary.append("Silent test fail for %s: Run separately!" % moduleResponse.fp)
        elif moduleResponse.returnCode == 'UntrappedException':
            otherSummary.append("Untrapped Exception for unknown module: %s" % moduleResponse.fp)
        elif moduleResponse.returnCode == 'TrappedException':
            otherSummary.append("Trapped Exception for module %s, at %s: %s" %
                                (moduleResponse.moduleName,
                                  moduleResponse.fp,
                                  moduleResponse.testRunner))
        elif moduleResponse.returnCode == 'LargeException':
            otherSummary.append("Large Exception for file %s: %s" %
                                (moduleResponse.fp, moduleResponse.testResult))
        elif moduleResponse.returnCode == 'ImportError':
            otherSummary.append("Import Error for %s" % moduleResponse.fp)
        elif moduleResponse.returnCode == 'NotInTree':
            if moduleResponse.moduleName == "":
                otherSummary.append("Not in Tree Error: %s " % moduleResponse.moduleName)
        elif moduleResponse.returnCode == 'TestsRun':
            totalTests += moduleResponse.testsRun
            if moduleResponse.success:
                successSummary.append("%s successfully ran %d tests in %d seconds"
                                      % (moduleResponse.moduleName,
                                         moduleResponse.testsRun,
                                         moduleResponse.runTime))
            else:
                errorsList = moduleResponse.errors
                # not the original errors list! see pickle note above
                failuresList = moduleResponse.failures
                errorsFoundSummary.append(
                    "\n-----------------------------\n"
                    + "%s had %d ERRORS and %d FAILURES in %d tests after %d seconds:\n" %
                    (moduleResponse.moduleName, len(errorsList),
                       len(failuresList), moduleResponse.testsRun, moduleResponse.runTime)
                    + "-----------------------------\n")

                for e in errorsList:
                    outStr += e + "\n"
                    errorsFoundSummary.append('%s' % (e))
                for f in failuresList:
                    outStr += f + "\n"
                    errorsFoundSummary.append('%s' % (f))
        else:
            otherSummary.append("Unknown return code %s" % moduleResponse)


    outStr += "\n\n---------------SUMMARY---------------------------------------------------\n"
    for l in skippedSummary:
        outStr += l + "\n"
    for l in successSummary:
        outStr += l + "\n"
    for l in otherSummary:
        outStr += l + "\n"
    for l in errorsFoundSummary:
        outStr += l + "\n"
    outStr += "-------------------------------------------------------------------------\n"
    elapsedTime = time.time() - timeStart
    outStr += "Ran %d tests in %.4f seconds\n" % (totalTests, elapsedTime)
    sys.stdout.flush()
    print(outStr)
    sys.stdout.flush()

    import datetime
    lastResults = os.path.join(tempfile.gettempdir(), 'lastResults.txt')
    with open(lastResults, 'w') as f:
        f.write(outStr)
        f.write("Run at " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    print("Results at " + lastResults)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-


def test_memory(test_function):
    try:
        import guppy
    except ImportError:
        raise ImportError("memoryUsage.py requires guppy")

    hp = guppy.hpy()
    hp.setrelheap()
    test_function()
    h = hp.heap()
    return h


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         commonTest.py
# Purpose:      Things common to testing
#
# Authors:      Christopher Ariza
#               Michael Scott Cuthbert
#
# Copyright:    Copyright © 2009-18 MIT DH Lab, Michael Scott Cuthbert
#               Forked from music21, Michael Scott Cuthbert
# License:      BSD, see license.txt
# ------------------------------------------------------------------------------
'''
Things that are common to testing...
'''
import doctest
import inspect
import os
import pathlib
import types
import warnings

import importlib
import unittest.runner
from unittest.signals import registerResult

from . import common


def default_doctest_suite(module_name=None, globs=None):
    if globs is True:
        globs = __import__(common.source_package_name()).__dict__.copy()
    elif globs in (False, None):
        globs = {}

    docTestOptions = (doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    kwArgs = {
              'globs': globs,
              'optionflags': docTestOptions,
              }
    # in case there are any tests here, get a suite to load up later
    if module_name is not None:
        s1 = doctest.DocTestSuite(module_name, **kwArgs)
    else:
        s1 = doctest.DocTestSuite(**kwArgs)
    return s1


# from testRunner...
# more silent type...
class ProjectTestRunner(unittest.runner.TextTestRunner):
    def run(self, test):
        "Run the given test case or test suite."
        result = self._makeResult()
        registerResult(result)
        result.failfast = self.failfast
        result.buffer = self.buffer
        with warnings.catch_warnings():
            if hasattr(self, 'warnings') and self.warnings:
                # if self.warnings is set, use it to filter all the warnings
                warnings.simplefilter(self.warnings)
                # if the filter is 'default' or 'always', special-case the
                # warnings from the deprecated unittest methods to show them
                # no more than once per module, because they can be fairly
                # noisy.  The -Wd and -Wa flags can be used to bypass this
                # only when self.warnings is None.
                if self.warnings in ['default', 'always']:
                    warnings.filterwarnings('module',
                            category=DeprecationWarning,
                            message=r'Please use assert\w+ instead.')
            # startTime = time.time()
            startTestRun = getattr(result, 'startTestRun', None)
            if startTestRun is not None:
                startTestRun()
            try:
                test(result)
            finally:
                stopTestRun = getattr(result, 'stopTestRun', None)
                if stopTestRun is not None:
                    stopTestRun()
            # stopTime = time.time()
        # timeTaken = stopTime - startTime
        result.printErrors()

        expectedFails = unexpectedSuccesses = skipped = 0
        try:
            results = map(len, (result.expectedFailures,
                                result.unexpectedSuccesses,
                                result.skipped))
        except AttributeError:
            pass
        else:
            expectedFails, unexpectedSuccesses, skipped = results

        infos = []
        if not result.wasSuccessful():
            self.stream.write("FAILED")
            failed, errored = len(result.failures), len(result.errors)
            if failed:
                infos.append("failures=%d" % failed)
            if errored:
                infos.append("errors=%d" % errored)
        else:
            pass
        if skipped:
            infos.append("skipped=%d" % skipped)
        if expectedFails:
            infos.append("expected failures=%d" % expectedFails)
        if unexpectedSuccesses:
            infos.append("unexpected successes=%d" % unexpectedSuccesses)
        if infos:
            self.stream.writeln(" (%s)" % (", ".join(infos),))
        else:
            pass
        return result


# ------------------------------------------------------------------------------
class ModuleGather:
    r'''
    Utility class for gathering and importing all modules in the project
    package. Puts them in self.modulePaths.

    >>> from dh_testers import commonTest
    >>> mg = commonTest.ModuleGather(stack_level=0, useExtended=True)
    >>> #_DOCS_SHOW print mg.modulePaths[0]
    D:\Web\eclipse\music21base\music21\volume.py
    '''

    def __init__(self, *, start_module=None, useExtended=False,
                 autoWalk=True, stack_level=None):
        if start_module is None:
            frame_record = common.get_first_external_stackframe()
            if frame_record is None and stack_level is None:
                stack_level = 1

            if stack_level is not None:
                frame_record = inspect.stack()[stack_level]

            mod_name = common.likely_python_module(frame_record.filename)
            start_module = __import__(mod_name)
        dirParent = pathlib.Path(start_module.__file__).parent

        # do not store start_module, since modules can't be pickled
        # self.start_module = start_module
        self.start_module_name = start_module.__name__
        self.dirParent = dirParent
        self.useExtended = useExtended
        self.modulePaths = []
        self.moduleSkip = []
        self.moduleSkipExtended = self.moduleSkip + []
        # run these first...
        self.slowModules = []

        # skip any path that contains this string
        self.pathSkip = []
        self.pathSkipExtended = self.pathSkip + []

        self.moduleSkip = [x.replace('/', os.sep) for x in self.moduleSkip]
        self.moduleSkipExtended = [x.replace('/', os.sep)
                                    for x in self.moduleSkipExtended]
        self.pathSkip = [x.replace('/', os.sep) for x in self.pathSkip]
        self.pathSkipExtended = [x.replace('/', os.sep)
                                    for x in self.pathSkipExtended]
        self.slowModules = [x.replace('/', os.sep) for x in self.slowModules]

        # search on init
        if autoWalk:
            self.walk()

    def _visitFunc(self, args, dirname, names):
        '''
        append all module paths from _walk() to self.modulePaths.
        Utility function called from os.walk()
        '''
        for fileName in names:
            if fileName.endswith('py'):
                fp = os.path.join(dirname, fileName)
                if not os.path.isdir(fp):
                    self.modulePaths.append(fp)

    def walk(self):
        '''
        Get all the modules in reverse order, storing them in self.modulePaths
        '''
        def manyCoreSortFunc(name):
            '''
            for many core systems, like the MacPro, running slowest modules first
            helps there be fewer idle cores later
            '''
            name = name[len(str(self.dirParent)) + 1:]
            name = name.replace('.py', '')
            return (name in self.slowModules, name)

        # the results of this are stored in self.curFiles, self.dirList
        for dirpath, unused_dirnames, filenames in os.walk(self.dirParent):
            self._visitFunc(None, dirpath, filenames)

        if common.cpus() > 4:# @UndefinedVariable
            self.modulePaths.sort(key=manyCoreSortFunc)
        else:
            self.modulePaths.sort()

        self.modulePaths.reverse()

    def _getName(self, fp):
        r'''
        Given full file pathlib.Path, find a name for the module
        with _ as the separator.

        >>> from dh_testers import commonTest
        >>> mg = commonTest.ModuleGather(stack_level=0)
        >>> #_DOCS_SHOW mg._getName(r'D:\music21base\music21\chord.py')
        'chord'
        '''
        fn = fp.replace(str(self.dirParent), '') # remove parent
        if fn.startswith(os.sep):
            fn = fn[1:]
        fn = fn.replace(os.sep, '_') # replace w/ _
        fn = fn.replace('.py', '')
        return fn

    def _getNamePeriod(self, fp, *, add_module_name=False):
        r'''
        Given full file path, find a name for the module with . as the separator.

        >>> from dh_testers import commonTest
        >>> mg = commonTest.ModuleGather(stack_level=0)
        >>> name = '/Users/cuthbert/git/music21base/music21/features/native.py'
        >>> #_DOCS_SHOW mg._getNamePeriod(name)
        'features.native'
        '''
        if not isinstance(fp, str):
            fp = str(fp)

        fn = fp.replace(str(self.dirParent), '') # remove parent
        parts = [x for x in fn.split(os.sep) if x]
        if parts[-1] == '__init__.py':
            parts.pop()
        fn = '.'.join(parts) # replace w/ period

        fn = fn.replace('.py', '')
        if add_module_name and fn:
            fn = add_module_name + '.' + fn
        elif add_module_name:
            fn = add_module_name


        return fn

    def load(self):
        '''
        Return a list of module objects that are not in the skip list.

        N.B. the list is a list of actual module objects not names,
        therefore cannot be pickled.
        '''
        modules = []
        for fp in self.modulePaths:
            moduleObject = self.getModule(fp)
            if moduleObject is not None:
                modules.append(moduleObject)
        return modules

    def getModule(self, fp):
        '''
        gets one module object from the file path
        '''
        skip = False
        ms = self.moduleSkip
        if self.useExtended:
            ms = self.moduleSkipExtended

        for fnSkip in ms:
            if fp.endswith(fnSkip):
                skip = True
                break
        if skip:
            return None

        ps = self.pathSkip
        if self.useExtended:
            ps = self.pathSkipExtended


        for dirSkip in ps:
            if dirSkip in fp:
                skip = True
                break
        if skip:
            return None

        name = common.likely_python_module(fp)
        # for importlib
        # name = self._getNamePeriod(fp, add_module_name='music21')

        # print(name, os.path.dirname(fp))
        try:
            with warnings.catch_warnings():
                # warnings.simplefilter('ignore', RuntimeWarning)
                # importlib is messing with coverage...
                mod = importlib.import_module(name)
                # mod = importlib.import_module(name)
        except Exception as excp: # pylint: disable=broad-except
            warnings.warn('failed import: ' + name + ' at ' + str(fp) + '\n'
                + '\tEXCEPTION:' + str(excp).strip())
            return None

        return mod

    def get_module_without_imp(self, fp, start_module=None):
        '''
        gets one module object from the file path without using imp

        start_module is the actual module object for the main project,
        such as "dhFrame" or "gender_novels" or "music21".  If None,
        it will try to import it.
        '''
        if start_module is None:
            start_module = common.import_main_module()
        skip = False
        for fnSkip in self.moduleSkip:
            if fp.endswith(fnSkip):
                skip = True
                break
        if skip:
            return "skip"

        for dirSkip in self.pathSkip:
            dirSkipSlash = os.sep + dirSkip + os.sep
            if dirSkipSlash in fp:
                skip = True
                break
        if skip:
            return "skip"
        moduleName = self._getNamePeriod(fp)
        moduleNames = moduleName.split('.')
        currentModule = start_module
        # print(currentModule, moduleName, fp)

        for thisName in moduleNames:
            if not thisName.strip():
                continue
            if hasattr(currentModule, thisName):
                currentModule = object.__getattribute__(currentModule, thisName)
                if not isinstance(currentModule, types.ModuleType):
                    return "notInTree"
            else:
                return "notInTree"
        mod = currentModule

        print('starting ' + moduleName)
        return mod


if __name__ == '__main__':
    from .testRunner import main_test
    main_test()



# -*- coding: utf-8 -*-

import os
import io

if __name__ == '__main__':
    directory = os.getcwd()

    for root, dirs, files in os.walk(directory):
        for f in files:
            if f.endswith('.py') is not True:
                continue
            fullf = root + os.sep + f
            if 'ext' in root:
                continue
            with io.open(fullf, encoding='latin-1') as fh:
                data = fh.read()
                head = data[0:200]
                if 'utf-8' not in head:
                    print(fullf, ' is not utf-8 compliant')



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         testRunner.py
# Purpose:      testing suite
#
# Authors:      Michael Scott Cuthbert
#               Christopher Ariza
#
# Copyright:    Copyright © 2018 MIT, Digital Humanities Lab
#               Forked from music21,
#               Copyright © 2006-2016 Michael Scott Cuthbert and the music21
#               Project
# License:      BSD, see license.txt
# -----------------------------------------------------------------------------
'''
The testRunner module contains the all important "main_test" function
that runs tests in a given module.
'''
import doctest
import inspect
import re
import sys
import unittest

# ALL_OUTPUT = []

###### test related functions
from . import common


def add_doc_attr_tests_to_suite(suite,
                           moduleVariableLists,
                           outerFilename=None,
                           globs=False,
                           optionflags=(doctest.ELLIPSIS
                                        | doctest.NORMALIZE_WHITESPACE
                                        )):
    '''
    takes a suite, such as a doctest.DocTestSuite and the list of variables
    in a module and adds from those classes that have a _DOC_ATTR dictionary
    (which documents the properties in the class) any doctests to the suite.
    '''
    dtp = doctest.DocTestParser()
    if globs is False:
        globs = __import__(common.source_package_name()).__dict__.copy()

    elif globs is None:
        globs = {}

    for lvk in moduleVariableLists:
        if not (inspect.isclass(lvk)):
            continue
        docattr = getattr(lvk, '_DOC_ATTR', None)
        if docattr is None:
            continue
        for dockey in docattr:
            documentation = docattr[dockey]
            # print(documentation)
            dt = dtp.get_doctest(documentation, globs, dockey, outerFilename, 0)
            if not dt.examples:
                continue
            dtc = doctest.DocTestCase(dt,
                                      optionflags=optionflags,
                                      )
            # print(dtc)
            suite.addTest(dtc)


def fixDoctests(doctestSuite):
    r'''
    Fix doctests so that adderesses are sanitized, and perhaps a few others.
    '''
    for dtc in doctestSuite: # Suite to DocTestCase -- undocumented.
        if not hasattr(dtc, '_dt_test'):
            continue

        dt = dtc._dt_test # DocTest
        for example in dt.examples: # fix Traceback exception differences Py2 to Py3
            example.want = stripAddresses(example.want, '0x...')


ADDRESS = re.compile('0x[0-9A-Fa-f]+')


def stripAddresses(textString, replacement="ADDRESS"):
    '''
    Function that changes all memory addresses (pointers) in the given
    textString with (replacement).  This is useful for testing
    that a function gives an expected result even if the result
    contains references to memory locations.  So for instance:

    >>> from dh_testers import testRunner
    >>> testRunner.stripAddresses("{0.0} <BlahBlah object at 0x02A87AD0>")
    '{0.0} <BlahBlah object at ADDRESS>'

    while this is left alone:

    >>> testRunner.stripAddresses("{0.0} <music21.humdrum.MiscTandem *>I humdrum control>")
    '{0.0} <music21.humdrum.MiscTandem *>I humdrum control>'


    For doctests, can strip to '...' to make it work fine with doctest.ELLIPSIS

    >>> testRunner.stripAddresses(
    ...     "{0.0} <BlahBlah object at 0x102a0ff10>", '0x...')
    '{0.0} <BlahBlah object at 0x...>'

    :rtype: str
    '''
    return ADDRESS.sub(replacement, textString)


# ------------------------------------------------------------------------------

def main_test(*testClasses,
              run_test=False,
              default_import=True,
              skip_doctest=False,
              verbose=False,
              module_relative=False,
              import_plus_relative=False,
              display_names=False,
              only_doctest=False,
              fail_fast=True):
    '''
    Takes as its arguments modules
    (or a string 'noDocTest' or 'verbose')
    and runs all of these modules through a unittest suite

    Unless 'noDocTest' is passed as a module, a docTest
    is also performed on `__main__`, hence the name "main_test".

    If 'moduleRelative' (a string) is passed as a module, then
    global variables are preserved.

    Run example (put at end of your modules):

    ::

        import unittest
        class Test(unittest.TestCase):
            def testHello(self):
                hello = "Hello"
                self.assertEqual("Hello", hello)

        if __name__ == '__main__':
            from dh_testers.testRunner import main_test
            main_test(Test)
    '''
    runAllTests = True

    # default -- is fail fast.
    if fail_fast:
        optionflags = (
            doctest.ELLIPSIS
            | doctest.NORMALIZE_WHITESPACE
            | doctest.REPORT_ONLY_FIRST_FAILURE
            )
    else:
        optionflags = (
            doctest.ELLIPSIS
            | doctest.NORMALIZE_WHITESPACE
            )

    globs = None
    # start with doc tests, then add unit tests
    if skip_doctest:
        # create a test suite for storage
        s1 = unittest.TestSuite()
    else:
        # create test suite derived from doc tests
        # here we use '__main__' instead of a module
        if not module_relative:
            if default_import:
                main_mod = common.import_main_module()
                if main_mod:
                    globs = main_mod.__dict__.copy()
                else:
                    globs = {}
            else:
                globs = {}
            if import_plus_relative:
                first_external = common.get_first_external_stackframe()
                if first_external is not None:
                    globs.update(first_external[0].f_globals)

        try:
            s1 = doctest.DocTestSuite(
                '__main__',
                globs=globs,
                optionflags=optionflags,
                )
        except ValueError as ve: # no docstrings
            print("Problem in docstrings [usually a missing r value before "
                  + "the quotes:] {0}".format(str(ve)))
            s1 = unittest.TestSuite()


    verbosity = 1
    if verbose:
        verbosity = 2 # this seems to hide most display

    if display_names:
        runAllTests = False

    runThisTest = None
    if len(sys.argv) == 2:
        arg = sys.argv[1].lower()
        if arg not in ('list', 'display', 'verbose', 'nodoctest'):
            # run a test directly named in this module
            runThisTest = sys.argv[1]
    if run_test:
        runThisTest = run_test

    # -f, --failfast
    if only_doctest:
        testClasses = [] # remove cases
    elif not testClasses:
        last_frame = common.get_first_external_stackframe()
        if last_frame is not None:
            last_frame_globals = last_frame[0].f_globals
            testClasses = []
            for k, v in last_frame_globals.items():
                if not inspect.isclass(v):
                    continue
                if (issubclass(v, unittest.TestCase)
                        and 'Slow' not in k
                        and 'External' not in k):
                    testClasses.append(v)

    for t in testClasses:
        if not isinstance(t, str):
            if display_names is True:
                for tName in unittest.defaultTestLoader.getTestCaseNames(t):
                    print('Unit Test Method: %s' % tName)
            if runThisTest is not None:
                tObj = t() # call class
                # search all names for case-insensitive match
                for name in dir(tObj):
                    if (name.lower() == runThisTest.lower()
                           or name.lower() == ('test' + runThisTest.lower())
                           or name.lower() == ('xtest' + runThisTest.lower())):
                        runThisTest = name
                        break
                if hasattr(tObj, runThisTest):
                    print('Running Named Test Method: %s' % runThisTest)
                    tObj.setUp()
                    getattr(tObj, runThisTest)()
                    runAllTests = False
                    break
                else:
                    print('Could not find named test method: %s, running all tests' % runThisTest)

            # normally operation collects all tests
            s2 = unittest.defaultTestLoader.loadTestsFromTestCase(t)
            s1.addTests(s2)

    ### Add _DOC_ATTR tests...
    if not skip_doctest:
        stacks = inspect.stack()
        if len(stacks) > 1:
            outerFrameTuple = stacks[1]
        else:
            outerFrameTuple = stacks[0]
        outerFrame = outerFrameTuple[0]
        outerFilename = outerFrameTuple[1]
        localVariables = list(outerFrame.f_locals.values())
        add_doc_attr_tests_to_suite(s1, localVariables, outerFilename, globs, optionflags)

    if runAllTests is True:
        fixDoctests(s1)

        runner = unittest.TextTestRunner()
        runner.verbosity = verbosity
        runner.run(s1)


if __name__ == '__main__':
    main_test()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# Name:         test.py
# Purpose:      Controller for all module tests in music21.
#
# Authors:      Christopher Ariza
#               Michael Scott Cuthbert
#
# Copyright:    Copyright © 2009-2012 Michael Scott Cuthbert and the music21 Project
# License:      LGPL or BSD, see license.txt
# ------------------------------------------------------------------------------
'''
Controller to run all module tests in the music21 folders.

Runs great, but slowly on multiprocessor systems.
'''

import doctest
import inspect
import sys
import unittest
import warnings

from . import common
from . import commonTest
from . import coverageProject
from . import testRunner


def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return str(msg) + '\n'


warnings.formatwarning = custom_formatwarning

_MOD = 'test.testSingleCoreAll'


# this is designed to be None for all but one system and a Coverage() object
# for one system.
cov = coverageProject.get_coverage()


def main(testGroup=('test',), limit=None, verbosity=2):
    '''
    Run all tests. Group can be test and external

    >>> print(None)
    None
    '''
    s1 = commonTest.default_doctest_suite(__name__)

    modGather = commonTest.ModuleGather()
    modules = modGather.load()

    # warnings.warn('looking for Test classes...\n')
    # look over each module and gather doc tests and unittests
    totalModules = 0
    sortMods = common.sort_modules(modules)
    # print(dir(sortMods[0]))
    main_module = common.import_main_module()
    if main_module:
        globs = main_module.__dict__.copy()
    else:
        globs = {}

    for moduleObject in sortMods:
        unitTestCases = []
        if limit is not None:
            if totalModules > limit:
                break
        totalModules += 1
        # get Test classes in module
        no_test_classes = True
        for global_name, global_var in moduleObject.__dict__.items():
            if not inspect.isclass(global_var):
                continue
            if issubclass(global_var, unittest.TestCase):
                if ('test' in testGroup
                        and 'External' not in global_name
                        and 'Slow' not in global_name):
                    unitTestCases.append(global_var)
                    no_test_classes = False

        if (no_test_classes
                and '__init__' not in moduleObject.__file__
                and 'test' not in moduleObject.__file__):
            warnings.warn('%s has no Test class' % moduleObject)

        if not hasattr(moduleObject, 'TestExternal'):
            pass
            # warnings.warn('%s has no TestExternal class\n' % module)
        else:
            if 'external' in testGroup or 'testExternal' in testGroup:
                unitTestCases.append(moduleObject.TestExternal)

        # for each Test class, load this into a suite
        for testCase in unitTestCases:
            s2 = unittest.defaultTestLoader.loadTestsFromTestCase(testCase)
            s1.addTests(s2)

        # add doc tests
        try:
            s3 = commonTest.default_doctest_suite(moduleObject, globs=globs)
            s1.addTests(s3)
        except ValueError:
            warnings.warn('%s cannot load Doctests' % moduleObject)
            continue

        allLocals = [getattr(moduleObject, x) for x in dir(moduleObject)]

        docTestOptions = (doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
        testRunner.add_doc_attr_tests_to_suite(
            s1,
            allLocals,
            outerFilename=moduleObject.__file__,
            globs=globs,
            optionflags=docTestOptions,
            # no checker here
            )

    testRunner.fixDoctests(s1)

    # warnings.warn('running Tests...\n')

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', RuntimeWarning)  # import modules...
        warnings.simplefilter('default', FutureWarning)  # a lot of these scipy->numpy
        runner = unittest.TextTestRunner(verbosity=verbosity)
        finalTestResults = runner.run(s1)

    coverageProject.stop_coverage(cov)

    if (finalTestResults.errors
            or finalTestResults.failures
            or finalTestResults.unexpectedSuccesses):
        returnCode = 1
    else:
        returnCode = 0

    return returnCode


def travisMain():
    # the main call for travis-ci tests.
    # exits with the returnCode
    returnCode = main(verbosity=1)
    exit(returnCode)


# ------------------------------------------------------------------------------
if __name__ == '__main__':
    # if optional command line arguments are given, assume they are
    # test group arguments
    if len(sys.argv) >= 2:
        unused_returnCode = main(sys.argv[1:])
    else:
        unused_returnCode = main()

# -----------------------------------------------------------------------------
# eof



# -*- coding: utf-8 -*-
import inspect
import multiprocessing
import os
import pathlib
import time


def likely_python_module(filename):
    '''
    Given a filename or Path, return the "likely" python module name.  That is, iterate the
    parent directories until it doesn't contain an __init__.py file.

    :rtype: str
    '''
    p = pathlib.Path(filename).resolve()
    paths = []
    if p.name != '__init__.py':
        paths.append(p.stem)
    while True:
        p = p.parent
        if not p:
            break
        if not p.is_dir():
            break

        inits = [f for f in p.iterdir() if f.name == '__init__.py']
        if not inits:
            break

        paths.append(p.stem)

    return '.'.join(reversed(paths))


def get_first_external_stackframe():
    '''
    Get the first FrameInfo that is not in the dh_testers project.

    Obviously, this can't be tested in the dh_testers project

    Returns None if none are found.

    :rtype: inspect.FrameInfo
    '''
    stack = inspect.stack()
    for frame in stack:
        fn = frame.filename
        if 'dh_testers' not in fn and 'python' not in fn:
            return frame
    return None


def source_file_path():
    '''
    Returns a pathlib.Path object for the directory of the package.

    :rtype: pathlib.Path
    '''
    package_name = source_package_name()
    package = __import__(package_name)
    return pathlib.Path(package.__file__).parent


def source_package_name():
    '''
    Returns the source package (the first module without a dot)
    '''
    package_frame = get_first_external_stackframe()
    if package_frame is None:
        return None
    package_name = likely_python_module(package_frame.filename)
    project_package = package_name.split('.')[0]
    return project_package


def import_main_module():
    spm = source_package_name()
    if spm is not None:
        return __import__(spm)


def sort_modules(moduleList):
    '''
    Sort a lost of imported module names such that most recently modified is
    first.  In ties, last accesstime is used then module name

    Will return a different order each time depending on the last mod time

    :rtype: list(str)
    '''
    sort = []
    modNameToMod = {}
    for mod in moduleList:
        modNameToMod[mod.__name__] = mod
        fp = mod.__file__ # returns the py or pyc file
        stat = os.stat(fp)
        lastmod = time.localtime(stat[8])
        asctime = time.asctime(lastmod)
        sort.append((lastmod, asctime, mod.__name__))
    sort.sort()
    sort.reverse()
    # just return module list
    outMods = [modNameToMod[modName] for lastmod, asctime, modName in sort]
    return outMods


def cpus():
    '''
    Returns the number of CPUs or if >= 3, one less (to leave something out for multiprocessing)
    '''
    cpuCount = multiprocessing.cpu_count() # @UndefinedVariable
    if cpuCount >= 3:
        return cpuCount - 1
    else:
        return cpuCount


if __name__ == '__main__':
    from .testRunner import main_test
    main_test()
